;; Generated package description from ebdb.el  -*- no-byte-compile: t -*-
(define-package "ebdb" "0.8.1" "Contact management package" '((emacs "25.1") (seq "2.15")) :keywords '("convenience" "mail") :authors '(("Eric Abrahamsen" . "eric@ericabrahamsen.net")) :maintainer '("Eric Abrahamsen" . "eric@ericabrahamsen.net") :url "https://github.com/girzel/ebdb")
